/*
 *****************************************************
 *
 *  SaVi by Robert Thurman (thurman@geom.umn.edu) and
 *          Patrick Worfolk (worfolk@alum.mit.edu).
 *
 *  Copyright (c) 1997 by The Geometry Center.
 *  This file is part of SaVi.  SaVi is free software;
 *  you can redistribute it and/or modify it only under
 *  the terms given in the file COPYRIGHT which you should
 *  have received along with this file.  SaVi may be
 *  obtained by emailing savi@geom.umn.edu.
 *
 *****************************************************
 *  
 * coverage_vis.c
 *
 * Coverage visualization routines.
 */
#include <tk.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "constants.h"
#include "stats_utils.h"
#include "utils.h"
#include "coverage_vis.h"
#include "orbit_utils.h"
#include "savi.h"

extern int tcl_script(char []);

static void grid_and_foreground_to_image(int projection, grid *g);
static void foreground_to_image(int projection, grid *g);
static void make_fat_dot(grid *g, int cindex, unsigned char *color);
static int new_foreground(int projection, grid *g, int force);
static void overlay_bitmap(unsigned int h, unsigned int w, unsigned char *b, 
			   const char *filename);
static void latlon_to_grid_index(int index[2], LatLon *pl, int projection, grid *g);

static Tk_PhotoImageBlock *image;
static unsigned char *foreground;

#define PIXELSIZE 3 /* 3 for RGB */
#define GT_DOT_FATNESS  3  /* Fatness of the ground track dot */


/* Colors for coverage intensity (in RGB values)  */
/* Any number of colors may be inserted.          */

static unsigned char colors[]={255,255,255,    /* No coverage     */ \
                               255,255,0,      /* Single coverage */ \
			       255,127,0,      /* Double coverage */ \
			       255,47,0,       /* Triple coverage */ \
			       255,0,0};      /* Four or more    */

#define num_colors sizeof(colors)/sizeof(unsigned char)/3

/* Color for earth outline and ground track */
static unsigned char outline_colors[]={255,255,255,  /* background */
                                       0,0,0,        /* earth outline */
				       255,0,0,   /* projected ground track */ 
				       255,0,0};  /* dynamic ground track */ 
#define num_outline_colors sizeof(outline_colors)/sizeof(unsigned char)/3

/* declarations for different coverage widget */
#ifdef COVERAGE2
#include "tkPlot.h"
#include "tcl_utils.h"

void update_display2(int projection, grid *g);
static TkPlot *plotPtr = NULL;
static int colors2[6];
#endif


/*
 * image_init.c
 * 
 * Set up the Tk image structure based on the coverage grid.
 *
 */
void
image_init(grid *g)
{
  unsigned int height, width, image_size;

  height=g->height;
  width =g->width;
  image_size = height*width;

  image = (Tk_PhotoImageBlock *) calloc(1, sizeof(Tk_PhotoImageBlock));
  image->pixelPtr =(unsigned char *)calloc(image_size*PIXELSIZE,
					   sizeof(unsigned char *));
  image->width = width;
  image->height = height;
  image->pitch = width*PIXELSIZE;
  image->pixelSize = PIXELSIZE;
  image->offset[0] = 0;
  image->offset[1] = 1;
  image->offset[2] = 2;

  foreground = (unsigned char *)calloc(image_size, sizeof(unsigned char *)); 

#ifdef COVERAGE2
  plotPtr = (TkPlot *) get_ClientData(".coverage.cmd.p");
  colors2[0] = Tk_PlotColorLookup("white");
  colors2[1] = Tk_PlotColorLookup("#ffff00");
  colors2[2] = Tk_PlotColorLookup("#ff7f00");
  colors2[3] = Tk_PlotColorLookup("#ff3f00");
  colors2[4] = Tk_PlotColorLookup("#ff1f00");
  colors2[5] = Tk_PlotColorLookup("black");
  Tk_PlotClear(plotPtr);
#endif
}

void
reset_foreground(int projection, grid *g)
{
  if (foreground != NULL)
    new_foreground(projection, g, TRUE);
}

/*
 * update_display
 *
 * Update the Tk coverage image, given a completed coverage grid.
 */
void
update_display(int projection, grid *g, int coverage_flag, int tracks_flag)
{
  Tk_PhotoHandle handle;
  char imagename[] = "im_coverage";

  handle = Tk_FindPhoto(imagename);
  if (handle != NULL)
    {
      /* Re-write the entire image, unless we are just displaying ground
       * tracks, and there's not a a new projection, in which case we can 
       * simply write the foreground (which includes ground tracks).  */

      if (new_foreground(projection, g, FALSE) || coverage_flag || 
	  !tracks_flag)
	grid_and_foreground_to_image(projection, g);
      else
	foreground_to_image(projection, g);
      Tk_PhotoPutBlock(handle, get_image(), 0, 0, image->width, image->height);
    }
  else
    {
      fprintf(stderr,"coverage_display: bad image name.\n");
    }

  set_coverage(projection,g); /* Display coverage percentages. */
}

/*
 * grid_to_image.c
 * 
 * Convert the intensity data in the coverage grid to colors in
 * the display image.
 *
 */
void
grid_to_image(grid *g)
{
  unsigned char *p;
  unsigned char *c;
  int ci, i;
  int image_size = g->height*g->width;

  p=image->pixelPtr;

  for (i=0; i<image_size; i++)
    {
      ci=g->data[i];
      if (ci >= num_colors) ci=num_colors-1;
      c=colors+3*ci;
      p[0]=c[0]; 
      p[1]=c[1]; 
      p[2]=c[2]; 
      p+=PIXELSIZE;
    }
  return;
}

/*
 * get_image.c
 *
 * Return pointer to the image structure.
 *
 */
Tk_PhotoImageBlock *
get_image()
{
  if (image == NULL)
    fprintf(stderr,"get_image: image not yet initialized.\n");
  return image;
}

/*
 * Clear the intensity array.
 */
void
clear_intensity(grid *g)
{
  int i;
  int image_size = g->height*g->width;

  for (i=0; i<image_size; i++)
    g->data[i]=0;
}

/*
 * set_coverage.c
 * 
 * Display total percent coverage.
 *
 */
void
set_coverage(int projection, grid *g)
{
  const char format1[] = "set coverage \"Current Coverage: %.1f%%\"";
  const char format2[] = "set mean_coverage \"Mean Coverage: %.1f%%\"";
  char cmd[sizeof(format1)+5];

  sprintf(cmd, format1, total_coverage(projection, g));
  tcl_script(cmd);
  sprintf(cmd, format2, mean_coverage(projection, g));
  tcl_script(cmd);
}

/*  
 * overlay_bitmap.c
 *
 * Takes a bitmap and lies it on top of our image array. 
 * h,w = height, width of the image array.
 */
static void
overlay_bitmap(unsigned int h, unsigned int w, unsigned char *b,
	       const char filename[])
{
  FILE *f;
  int pbmtype, width, height, i, j;
  int image_size = h*w;
  int c, mask;

  if (NULL == (f = fopen(filename, "r"))) return;

  forward_over_comments(f);
  fscanf(f, "P%d", &pbmtype);
  forward_over_comments(f);
  fscanf(f, "%d", &width);
  forward_over_comments(f);
  fscanf(f, "%d", &height);
  /* skip one char */
  c = getc(f);
  
  if ((pbmtype == 4) && (width == w) && (height == h))
    {
      for (i=0; i<image_size; i += 8)
	{
	  c = getc(f);
	  mask = 0x80;
	  for (j=0; j<8; j++)
	    {
	      if (c & mask) *(b++) = 1;
	      else b++;
	      mask = mask >> 1;
	    }
	}
    }
  else
    {
      fprintf(stderr,"Error in bitmap format: P%d %d %d\n", 
	      pbmtype, width, height);
    }

  fclose(f);
}



/*
 * new_foreground
 *
 * Initialize the foreground intensity image, including earth outline.
 * Returns TRUE if foreground must be re-drawn, FALSE otherwise.
 *
 */
static int 
new_foreground(int projection, grid *g, int force)
{
  static int current_proj = -1;
  SphericalCoordinates point;
  int left[2], right[2];
  unsigned int h, w, i;
  int image_size = g->width*g->height;

  if (!force && (projection == current_proj)) return(FALSE);
  current_proj = projection;

  memset(foreground, 0, (size_t) sizeof(int)*image_size);
    
  if (projection == SINUSOIDAL)
    {
      point.r = 1;
      point.theta = 0;
      point.phi = 0;
      
      h = g->height;
      w = g->width;

      for (i=0; i< h; i++)
	{
	  intensity_edges(left, right, &point, projection, g); 
	  foreground[(w*i+left[0])]=1;
	  foreground[(w*i+right[0])]=1;
	  point.phi += PI/(h-1);  /* Increment phi by one pixel */ 
	}
      overlay_bitmap(g->height, g->width, foreground, SINUSOIDAL_BITMAP_NAME);
    }
  else
    {
      overlay_bitmap(g->height, g->width, foreground, RECTANGULAR_BITMAP_NAME);
    }
  return(TRUE);
}



/*
 * Copy grid and foreground (land masses and ground tracks, if any) onto 
 * image.  Every image pixel is re-drawn.
 */
static void
grid_and_foreground_to_image(int projection, grid *g)
{
  int i, ci;
  unsigned char *c;
  unsigned char *f = foreground;
  unsigned char fg;
  unsigned char *p = image->pixelPtr;
  int image_size = g->width*g->height;

  for (i=0; i<image_size; i++)
    {
      fg = *(f++);
      if (fg > 0)
	{
	  c = outline_colors+3*fg;
#ifdef DEBUG
	  if (fg >= num_outline_colors) 
	    fprintf(stderr, "foreground_to_image: foreground index out of range.\n");
#endif
	  /* Foreground index of 1 indicates a map outline, 
	     2 indicates a projected ground track,
	     3 indicates dynamically-drawn ground track, which we want
	     to be drawn with a fat dot. */

	  if (fg >= 3) 
	    make_fat_dot(g,i,c);
	  else 
	    {
	      p[0] = c[0];
	      p[1] = c[1];
	      p[2] = c[2];
	    }
	}
      else
	{
	  ci = g->data[i];
	  if (ci >= num_colors) ci=num_colors-1;
	  c = colors+3*ci;
	  p[0] = c[0];
	  p[1] = c[1];
	  p[2] = c[2];
	}
      p += PIXELSIZE;
    }
}
/*
 * Copy foreground (land masses and ground tracks, if any) only onto image.
 * The routine re-draws an image pixel only for corresponding foreground array 
 * elements which are non-zero.
 */
static void
foreground_to_image(int projection, grid *g)
{
  int i;
  unsigned char *c;
  unsigned char *f = foreground;
  unsigned char fg;
  unsigned char *p = image->pixelPtr;
  int image_size = g->width*g->height;

  for (i=0; i<image_size; i++)
    {
      fg = *(f++);
      if (fg > 0)
	{
	  c = outline_colors+3*fg;
#ifdef DEBUG
	  if (fg >= num_outline_colors) 
	    fprintf(stderr, "foreground_to_image: foreground index out of range.\n");
#endif
	  /* Foreground index of 1 indicates a map outline, 
	     2 indicates a projected ground track,
	     3 indicates dynamically-drawn ground track, which we want
	     to be drawn with a fat dot. */

	  if (fg >= 3) 
	    make_fat_dot(g,i,c);
	  else 
	    {
	      p[0] = c[0];
	      p[1] = c[1];
	      p[2] = c[2];
	    }
	}
      p += PIXELSIZE;
    }
}

/*
 *  make_fat_dot
 *
 *  Make a fat dot on the image. Fat dot is actually a square, whose
 *  lower right-hand corner is the base point from the grid.
 *
 *     g      -- coverage grid
 *     index  -- index of base point in the grid
 *     color  -- pointer to color array for the dot
 *
 *  The size of the fat dot is determined by the parameter GT_DOT_FATNESS.
 */
static void
make_fat_dot(grid *g, int index, unsigned char *color)
{
  unsigned char *p = image->pixelPtr;
  int i_row, i_col, f_row, f_col, i, j;
  int w = g->width;
  int h = g->height;

  /* Row and column indices of the base point. */
  f_row = index/w;     /* ranges from 0 to height-1 */
  f_col = index-f_row*w; /* ranges from 0 to width-1 */

  i_row = f_row-GT_DOT_FATNESS+1;    /* Indices of the initial row */
  i_col = f_col-GT_DOT_FATNESS+1;    /* and column of the fat dot. */
  
  p += (i_row*w + i_col)*PIXELSIZE;

  for (i = i_row; i<=f_row; i++)
    {
      for (j = i_col; j<= f_col; j++)
	{
	  if (i < h && i > -1 && j < w && j > -1 )
	    {
	      p[0] = color[0];        /* color the pixel */
	      p[1] = color[1];
	      p[2] = color[2];
	    }
	  p += PIXELSIZE;
	}
      p += (w-GT_DOT_FATNESS)*PIXELSIZE;
    }
}

#ifdef COVERAGE2
void
update_display2(int projection, grid *g)
{
  int *data = g->data;

  if (plotPtr == NULL) return;

  new_foreground(projection, g);

  Tk_PlotGridForeground(plotPtr, g, foreground, colors2, 5);
  Tk_PlotQueueRefresh(plotPtr);

  return;
}
#endif


/*
 * tracks_to_foreground
 *
 * routine to put a dot on the foreground map which corresponds
 * to the point over which the satellite is.  For every satellite
 * in the list.
 */
void
tracks_to_foreground(const Satellite_list SL, int projection, grid *g,
		     int index, const CentralBody *pcb)
{
  Satellite_list sl=SL;

  while (sl)
    {
      track_to_foreground(sl->s, projection, g, index, pcb);
      sl = sl->next;
    }
}

/*
 * track_to_foreground
 *
 * routine to increment the foreground map at the point underneath
 * the satellite.  For a single satellite.
 *
 * index = color index to insert into the foreground array
 */
void
track_to_foreground(Satellite s, int projection, grid *g, int index,
		    const CentralBody *pcb)
{
  LatLon l;
  int coords[2];

  spherical_to_lat_lon(&l, &(s->x_S), s->t, pcb);
  latlon_to_grid_index(coords, &l, projection, g);
  foreground[g->width*coords[1]+coords[0]]=index;
}



/*
 * Given lat/lon coordinates of point on central body, return corresponding 
 * index into grid.  
 *        index[0]=column index, index[1]=row index.
 */
static void
latlon_to_grid_index(int index[2], LatLon *pl, int projection, grid *g)
{
  double proj[2];
  LatLon l;

/* The vertical center-line of the grid corresponds to longitude = 
 * LONGITUDE_CENTER_LINE, whereas both 2-d projections place longitude = 0 
 * along this line.  So shift the input longitude.
 */
  l.lat = pl->lat;
  l.lon = pl->lon - LONGITUDE_CENTER_LINE;

  if (l.lon < -180.0) l.lon += 360.0;
  else if (l.lon > 180.0) l.lon -= 360.0;

  if (projection == SINUSOIDAL)
    project_latlon_sinusoidal(proj, &l);
  else
    project_latlon_rectangular(proj, &l);


/* It's assumed that proj[0] ranges from -180 to 180, and
 *                   proj[1] ranges from -90 to 90,  no matter what. */
  index[0]=((int)((proj[0]+180.0)*(g->width)/360.0));
  index[1]=((int)(proj[1]*(g->height-1)/180.0));
}


/*
 * write_grid
 *
 * writes grid to file in ??? format
 *
 */
void
write_coverage_image(FILE *out)
{
  int i;
  int image_size;
  int pixel_size;
  unsigned char *data;

  /* get pointer to image and return if it doesn't eist */
  Tk_PhotoImageBlock *image = get_image();
  if (image == NULL) return;

  image_size = image->width*image->height;
  pixel_size = image->pixelSize;
  data = image->pixelPtr;

  /* output header */
  fprintf(out, "P6 %d %d 255\n", image->width, image->height);

  for (i=0; i<image_size; i++) {
    fputc(data[0], out);
    fputc(data[1], out);
    fputc(data[2], out);
    data += pixel_size;
  }

}


