/*
 *****************************************************
 *
 *  SaVi by Robert Thurman (thurman@geom.umn.edu) and
 *          Patrick Worfolk (worfolk@alum.mit.edu).
 *
 *  Copyright (c) 1997 by The Geometry Center.
 *  This file is part of SaVi.  SaVi is free software;
 *  you can redistribute it and/or modify it only under
 *  the terms given in the file COPYRIGHT which you should
 *  have received along with this file.  SaVi may be
 *  obtained by emailing savi@geom.umn.edu.
 *
 *****************************************************
 *  
 * orbit_utils.h
 *
 */

#ifndef _ORBIT_UTILS_H_
#define _ORBIT_UTILS_H_

#include "utils.h"

/*
 * Define a struct for orbital elements
 *
 * Note: we use time of periapsis whereas it is also common to use
 *       mean anomaly, MA.  To convert one to the other use the formula:
 *               T = -MA/360 * 2Pi * a^(3/2) / sqrt(mu)
 *       Where 2Pi * a^(3/2) / sqrt(mu) is of course the period of the orbit.
 */
typedef struct OrbitalElements {
  double 
    a,              /* length of semi-major axis   */
    e,              /* eccentricity                */
    i,              /* inclination                 */
    Omega,          /* longitude of ascending node */
    omega,          /* argument of periapsis       */
    T;              /* time of periapsis           */
} OrbitalElements;


/*
 * Define a struct to hold information about a central body
 *
 */
typedef struct CentralBody {
  double
    radius,
    rotation_rate,
    mu,
    J2,
    orbital_period,
    tilt;
  
} CentralBody;

void oe_time_to_geocentric(CartesianCoordinates *px, double t, 
			     OrbitalElements *poe, const CentralBody *pcb);
void oe_time_to_geocentric_spherical(SphericalCoordinates *px, double t, 
			     OrbitalElements *poe, const CentralBody *pcb);
void spherical_to_transform(double T[4][4], SphericalCoordinates *px,
			    double scale);
void spherical_to_cartesian(CartesianCoordinates *px, 
			    SphericalCoordinates *py);
void oe_time_J0_to_oe(OrbitalElements *final, OrbitalElements *initial,
		      double t, const CentralBody *pcb);
void oe_time_J2_to_oe(OrbitalElements *final, OrbitalElements *initial,
		      double t, const CentralBody *pcb);
void oe_time_to_oe(OrbitalElements *final, OrbitalElements *initial, double t,
		   const CentralBody *pcb);
void oe_to_geocentric(CartesianCoordinates *px, OrbitalElements *poe,
		      const CentralBody *pcb);

double oe_to_period(OrbitalElements *poe, const CentralBody *pcb);

void perifocal_to_geocentric(CartesianCoordinates *, CartesianCoordinates *,
			     double, double, double);

void cartesian_to_spherical(SphericalCoordinates *px,
			    CartesianCoordinates *py);
void geocentric_to_horizon(CartesianCoordinates *origin, 
			   CartesianCoordinates *target, 
			   HorizonCoordinates *u); 
void horizon_to_geocentric(HorizonCoordinates *u, 
			   CartesianCoordinates *origin, 
			   CartesianCoordinates *target);
void horizon_plane (CartesianCoordinates *origin,
		    CartesianCoordinates *e, CartesianCoordinates *n);
void lat_lon_to_spherical (double lat, double lon, double t,
			   const CentralBody *pcb, SphericalCoordinates *px);
void lat_lon_to_cartesian (double lat, double lon, double t,
			   const CentralBody *pcb, CartesianCoordinates *px);
void spherical_to_lat_lon(LatLon *pl, SphericalCoordinates *px, double t,
		     const CentralBody *pcb);
double eccentric_anomaly_to_true_anomaly (double , double );
double time_to_true_anomaly (double t, const CentralBody *pcb, double a, double e,
		                    double T);
int sat_to_fisheye (CartesianCoordinates *px, double lat, double lon,
		    const CentralBody *pcb, double t, double sky[2]);

void oe_info(FILE *, OrbitalElements *poe, const CentralBody *pcb);

double oe_to_nodal_precession(OrbitalElements *poe, const CentralBody *pcb);
double oe_to_apsidal_rotation(OrbitalElements *poe, const CentralBody *pcb);
double oe_to_apoapsis_altitude(OrbitalElements *poe, const CentralBody *pcb);
double oe_to_periapsis_altitude(OrbitalElements *poe, const CentralBody *pcb);

double mean_anomaly(double t, double mu, double a, double T);

#endif
/* !_ORBIT_UTILS_H_ */
