/*
 *****************************************************
 *
 *  SaVi by Robert Thurman (thurman@geom.umn.edu) and
 *          Patrick Worfolk (worfolk@alum.mit.edu).
 *
 *  Copyright (c) 1997 Patrick Worfolk
 *  All rights reserved.
 *
 *****************************************************
 *  
 * links.c
 *
 * routines for inter-satellite links
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "globals.h"
#include "constants.h"
#include "gv_utils.h"
#include "sats.h"
#include "orbit_utils.h"
#include "savi.h"

static int links_on_flag = FALSE;
static int links_n = 0;
static int links_geom_exists = FALSE;
static int links_geom_needs_updating = FALSE;

static void
link_write_geom(const Satellite s, const Constellation *pconstellation);
static void
link_output(FILE *fp, Satellite s, const Constellation *pconstellation,
	    int n);

typedef enum {
  SAME_PLANE_AHEAD_2=0,
  SAME_PLANE_AHEAD_1,
  SAME_PLANE_BEHIND_1,
  SAME_PLANE_BEHIND_2,
  NEIGHBOR_RIGHT_2,
  NEIGHBOR_RIGHT_1,
  NEIGHBOR_LEFT_1,
  NEIGHBOR_LEFT_2,
  N_LINK_TYPES} Link_Types;

static int link_flags[N_LINK_TYPES] = {OFF, OFF, OFF, OFF, OFF, OFF, OFF, OFF};

/*
 * links_on_cmd()
 *
 * Routine to display links
 */
char *
links_on_cmd(int argc, char *argv[])
{
  int i;
  Constellation *pconstellation = get_constellation();
  Satellite_list sl = pconstellation->satellites;

  if (pconstellation->plan.n_planes <= 0) {
    fprintf(stderr, "Orbital plane structure has not been defined.\n");
    return EMPTY_str;
  }

  /* decode which links are needed */
  links_n = 0;
  for (i=0; i<N_LINK_TYPES; i++) {
    if (atoi(argv[2+i]) == 1) {
      link_flags[i]=ON;
      links_n++;
    }
    else link_flags[i]=OFF;
  }

  /* if no links needed then interpret this as a links off command */
  if (links_n == 0)
    return links_off_cmd(argc, argv);


  /* now update the links structure - we don't care if it existed before
     since we may have changed the active links
  */


  fprintf(stderr, "links on (%d)\n", links_n);
  links_on_flag = TRUE;
  links_geom_needs_updating = TRUE;

  gv_begin();
  while (sl)
    {
      link_display(sl->s, pconstellation);
      sl = sl->next;
    }
  gv_end();

  return EMPTY_str;
}

/*
 * links_off_cmd()
 *
 * Routine to turn off display of links
 */
char *
links_off_cmd(int argc, char *argv[])
{
  Satellite_list sl = get_constellation()->satellites;

  if (FALSE == links_on_flag) return EMPTY_str;

  fprintf(stderr, "links off...\n");
  gv_begin();
  links_gv_delete();
  while (sl)
    {
      link_delete(sl->s);
      sl = sl->next;
    }
  links_geom_needs_updating = FALSE;
  links_on_flag = FALSE;
  gv_end();

  return EMPTY_str;
}


char *
links_set_cmd(int argc, char *argv[])
{
  Constellation *pconstellation = get_constellation();
  Satellite s;

  int n = atoi(argv[2]);
  unsigned int flag = atoi(argv[3]);

  s = constellation_search(pconstellation, n);
  if (s) {
    if (flag) s->can_display_links = 1;
    else s->can_display_links = 0;
    links_geom_needs_updating = TRUE;
  }

  return EMPTY_str;
}

/*
 * links_rebuild()
 *
 * Routine to resize all the links currently built
 */
void
links_rebuild()
{
  Constellation *pconstellation = get_constellation();
  Satellite_list sl = pconstellation->satellites;

  if (FALSE == links_on_flag) return;

  gv_begin();
  while (sl) {
    link_display(sl->s, pconstellation);
    sl = sl->next;
  }
  gv_end();
  
  links_geom_needs_updating = FALSE;
}


/*
 * links_gv_delete
 *
 * routine to delete geomview links object
 */
void
links_gv_delete()
{
  if (TRUE == links_geom_exists) {
    gv_delete_geom("Links");
    links_geom_exists = FALSE;
  }
}


/*
 * link_display
 *
 * Takes a satellite and displays its link
 */
void
link_display(Satellite s, const Constellation *pconstellation)
{
  if (FALSE == links_on_flag) return;
  links_geom_needs_updating = TRUE;

  if (s->can_display_links)
    link_write_geom(s, pconstellation);
}




/*
 * link_delete
 *
 * Takes a satellite and deletes its link
 * ONLY if links_on_flag == TRUE
 */
void
link_delete(Satellite s)
{
  static char name[] = "link_XXXXXXXXXX";

  if (FALSE == links_on_flag) return;
  links_geom_needs_updating = TRUE;

  gv_begin();
  sprintf(name+5, "%d", s->id);
  gv_delete_handle(name);
  gv_end();
}



void
write_links_geom(const Constellation *pconstellation)
{
  Satellite s;
  Satellite_list sl = pconstellation->satellites;
  char buf[256];

  if ((links_on_flag) && (links_geom_needs_updating == TRUE))
    {
      /* write out new list of all satellites */
      gv_send("(geometry Links {LIST\n");
      while (NULL != sl)
	{
	  s = sl->s;
	  if (s->can_display_links) {
	    sprintf(buf, "{ :link_%d }\n", s->id);
	    gv_send(buf);
	  }
	  sl = sl->next;
	}
      gv_send("})\n");
      
      links_geom_exists = TRUE;
      links_geom_needs_updating = FALSE;
    }
}

/*
 * links_relocate
 *
 * Called to update the links when the satellites have all been moved
 *
 */
void
links_relocate(const Constellation *pconstellation)
{
  Satellite_list sl = pconstellation->satellites;

  if (FALSE == links_on_flag) return;
  
  while (sl) {
    if (sl->s->can_display_links)
      link_write_geom(sl->s, pconstellation);
    sl = sl->next;
  }

}

static void
link_write_geom(const Satellite s, const Constellation *pconstellation)
{
  char buf[256];
  CartesianCoordinates *px = &(s->x_C), *py;
  double r = pconstellation->pcb->radius;
  Satellite ss;
  int i;

  /* send header */
  gv_begin();
  sprintf(buf, "(read geometry {define link_%d { VECT %d %d 1\n",
	  s->id, links_n, 2*links_n);
  gv_send(buf);
  /* send points in each polyline */
  for (i=0; i<links_n; i++)
    gv_send(" 2");
  /* send colors in each polyline */
  gv_send("\n 1");
  for (i=1; i<links_n; i++)
    gv_send(" 0");
  gv_send("\n");

  /* write the points */
  if (link_flags[SAME_PLANE_AHEAD_2] == ON) {
    ss = inplane_neighbor(s, 2);
    if (ss) {
      py = &(ss->x_C);
    } else {
      py = px;
    }
    sprintf(buf, "%f %f %f %f %f %f\n",
	    px->x/r, px->y/r, px->z/r,
	    py->x/r, py->y/r, py->z/r);
    gv_send(buf);
  }
  if (link_flags[SAME_PLANE_AHEAD_1] == ON) {
    ss = inplane_neighbor(s, 1);
    if (ss) {
      py = &(ss->x_C);
    } else {
      py = px;
    }
    sprintf(buf, "%f %f %f %f %f %f\n",
	    px->x/r, px->y/r, px->z/r,
	    py->x/r, py->y/r, py->z/r);
    gv_send(buf);
  }
  if (link_flags[SAME_PLANE_BEHIND_1] == ON) {
    ss = inplane_neighbor(s, -1);
    if (ss) {
      py = &(ss->x_C);
    } else {
      py = px;
    }
    sprintf(buf, "%f %f %f %f %f %f\n",
	    px->x/r, px->y/r, px->z/r,
	    py->x/r, py->y/r, py->z/r);
    gv_send(buf);
  }
  if (link_flags[SAME_PLANE_BEHIND_2] == ON) {
    ss = inplane_neighbor(s, -2);
    if (ss) {
      py = &(ss->x_C);
    } else {
      py = px;
    }
    sprintf(buf, "%f %f %f %f %f %f\n",
	    px->x/r, px->y/r, px->z/r,
	    py->x/r, py->y/r, py->z/r);
    gv_send(buf);
  }
  if (link_flags[NEIGHBOR_RIGHT_2] == ON) {
    ss = outofplane_neighbor(s, 2);
    if (ss) {
      py = &(ss->x_C);
    } else {
      py = px;
    }
    sprintf(buf, "%f %f %f %f %f %f\n",
	    px->x/r, px->y/r, px->z/r,
	    py->x/r, py->y/r, py->z/r);
    gv_send(buf);
  }
  if (link_flags[NEIGHBOR_RIGHT_1] == ON) {
    ss = outofplane_neighbor(s, 1);
    if (ss) {
      py = &(ss->x_C);
    } else {
      py = px;
    }
    sprintf(buf, "%f %f %f %f %f %f\n",
	    px->x/r, px->y/r, px->z/r,
	    py->x/r, py->y/r, py->z/r);
    gv_send(buf);
  }
  if (link_flags[NEIGHBOR_LEFT_1] == ON) {
    ss = outofplane_neighbor(s, -1);
    if (ss) {
      py = &(ss->x_C);
    } else {
      py = px;
    }
    sprintf(buf, "%f %f %f %f %f %f\n",
	    px->x/r, px->y/r, px->z/r,
	    py->x/r, py->y/r, py->z/r);
    gv_send(buf);
  }
  if (link_flags[NEIGHBOR_LEFT_2] == ON) {
    ss = outofplane_neighbor(s, -2);
    if (ss) {
      py = &(ss->x_C);
    } else {
      py = px;
    }
    sprintf(buf, "%f %f %f %f %f %f\n",
	    px->x/r, px->y/r, px->z/r,
	    py->x/r, py->y/r, py->z/r);
    gv_send(buf);
  }


  /* write the colors and trailer */
  gv_send("0 0 1 1 } } )");
  gv_end();
}

/*
 * links_output
 *
 * Writes out links of all satellites in a constellation
 * with the can_display_links flag set
 *
 */
void
links_output(FILE *fp, const Constellation *pconstellation)
{
  Satellite_list sl = pconstellation->satellites;
  int n=1;

  while (sl) {
    if (sl->s->can_display_links) {
      fprintf(fp, "\tLINKS %d\n", n);
      link_output(fp, sl->s, pconstellation, n);
    }
    n++;
    sl = sl->next;
  }

}

/*
 * link_output
 *
 * outputs all links from a satellite
 *
 */
static void
link_output(FILE *fp, Satellite s, const Constellation *pconstellation,
	    int n)
{
  Satellite ss;

  if (link_flags[SAME_PLANE_AHEAD_2] == ON) {
    fprintf(fp, "\t\t 0+2\t");
    ss = inplane_neighbor(s, 2);
    if (ss) {
      fprintf(fp, "\n");
    } else {
      fprintf(fp, "link not available\n");
    }
  }
  if (link_flags[SAME_PLANE_AHEAD_1] == ON) {
    fprintf(fp, "\t\t 0+1\t");
    ss = inplane_neighbor(s, 1);
    if (ss) {
      fprintf(fp, "\n");
    } else {
      fprintf(fp, "link not available\n");
    }
  }
  if (link_flags[SAME_PLANE_BEHIND_1] == ON) {
    fprintf(fp, "\t\t 0-1\t");
    ss = inplane_neighbor(s, -1);
    if (ss) {
      fprintf(fp, "\n");
    } else {
      fprintf(fp, "link not available\n");
    }
  }
  if (link_flags[SAME_PLANE_BEHIND_2] == ON) {
    fprintf(fp, "\t\t 0-2\t");
    ss = inplane_neighbor(s, -2);
    if (ss) {
      fprintf(fp, "\n");
    } else {
      fprintf(fp, "link not available\n");
    }
  }
  if (link_flags[NEIGHBOR_RIGHT_2] == ON) {
    fprintf(fp, "\t\t+2\t");
    ss = outofplane_neighbor(s, 2);
    if (ss) {
      fprintf(fp, "\n");
    } else {
      fprintf(fp, "link not available\n");
    }
  }
  if (link_flags[NEIGHBOR_RIGHT_1] == ON) {
    fprintf(fp, "\t\t+1\t");
    ss = outofplane_neighbor(s, 1);
    if (ss) {
      fprintf(fp, "\n");
    } else {
      fprintf(fp, "link not available\n");
    }
  }
  if (link_flags[NEIGHBOR_LEFT_1] == ON) {
    fprintf(fp, "\t\t-1\t");
    ss = outofplane_neighbor(s, -1);
    if (ss) {
      fprintf(fp, "\n");
    } else {
      fprintf(fp, "link not available\n");
    }
  }
  if (link_flags[NEIGHBOR_LEFT_2] == ON) {
    fprintf(fp, "\t\t-2\t");
    ss = outofplane_neighbor(s, -2);
    if (ss) {
      fprintf(fp, "\n");
    } else {
      fprintf(fp, "link not available\n");
    }
  }
  
  /* also output the sun-boreline angle */
  fprintf(fp, "\t\tSUN boreline angle %f\n",
	  sun_angle(s, pconstellation->pcb));
}
