/*
 * plan.c
 *
 * routines to manipulate the declared structure of a constellation
 *
 */
#include <stdio.h>
#include <stdlib.h>

#include "Satellite.h"
#include "sats.h"
#include "savi.h"
#include "globals.h"
#include "utils.h"

static void
plan_find_satellite(Plan *pp, int id, int *plane, int *index);

char *
plan_nplanes_cmd(int argc, char *argv[])
{
  Constellation *pc = get_constellation();
  Plan *pp = &(pc->plan);
  int n = atoi(argv[2]);
  int i;
  
  plan_clear(pc);
  if (n<0) {
    /* error */
    error("NPLANES command needs a positive integer.");
    return EMPTY_str;
  }

  pp->n_planes = n;
  pp->sats_in_plane = (int *) calloc((unsigned int) n, sizeof(int));
  pp->sat_ids = (int **) calloc((unsigned int) n, sizeof(int *));
  for (i=0; i<n; i++) {
    pp->sats_in_plane[i] = 0;
    pp->sat_ids[i] = NULL;
  }

  return EMPTY_str;
}

char *
plan_plane_n_cmd(int argc, char *argv[])
{
  Constellation *pc = get_constellation();
  Plan *pp = &(pc->plan);
  int plane = atoi(argv[2]);
  int n = argc - 3;
  int i;

  if ( (plane<0) || (plane >= pp->n_planes)) {
    /* error - bad plane number */
    error("PLANE_N command specified an invalid plane.");
  }
  
  /* free up previous info */
  if (pp->sat_ids[plane])
    free(pp->sat_ids[plane]);

  /* make sure there are not a negative number of satellites in the plane */
  if (n<0) n=0;
  pp->sats_in_plane[plane] = n;
  pp->sat_ids[plane] = (int *) calloc((unsigned int) n, sizeof(int));
  for (i=0; i<n; i++)
    pp->sat_ids[plane][i] = atoi(argv[i+3]);

  return EMPTY_str;
}

void
plan_clear(Constellation *pc)
{
  Plan *pp = &(pc->plan);
  int i;

  if (pp->sat_ids) {
    for (i=0; i<pp->n_planes; i++)
      if (pp->sat_ids[i])
	free(pp->sat_ids[i]);
    free(pp->sat_ids);
    pp->sat_ids = NULL;
  }
  if (pp->sats_in_plane) {
    free(pp->sats_in_plane);
    pp->sats_in_plane = NULL;
  }
  pp->n_planes = 0;
}

char *
plan_dump_cmd(int argc, char *argv[])
{
  Constellation *pc = get_constellation();
  Plan *pp = &(pc->plan);
  int i, j;

  if (pp->n_planes == 0) {
    
    fprintf(stderr, "Constellation has no plan specified.\n");

  } else {

    fprintf(stderr, "Constellation is structured in %d planes.\n", pp->n_planes);
    for (i=0; i<pp->n_planes; i++) {
      fprintf(stderr, "  Plane %d: %d satellites ( ", i, pp->sats_in_plane[i]);
      for (j=0; j<pp->sats_in_plane[i]; j++)
	fprintf(stderr, "%d ", pp->sat_ids[i][j]);
      fprintf(stderr, ")\n");
    }

  }

  return EMPTY_str;
}


Satellite
inplane_neighbor(Satellite ps, int index)
{
  Constellation *pc = get_constellation();
  Plan *pp = &(pc->plan);
  int plane, idx, nidx;

  if (pp->n_planes == 0)
    return NULL;  /* no known structure */

  plan_find_satellite(pp, ps->id, &plane, &idx);

  /* satellite not in plan */
  if ((plane < 0) || (pp->sats_in_plane[plane] == 0))
    return NULL;

  nidx = idx + index;
  while (nidx < 0)
    nidx += pp->sats_in_plane[plane];
  nidx = nidx % pp->sats_in_plane[plane];

  return constellation_search(pc, pp->sat_ids[plane][nidx]);
}

Satellite
outofplane_neighbor(Satellite ps, int index)
{
  Constellation *pc = get_constellation();
  Plan *pp = &(pc->plan);
  Satellite closest, current;
  double range;
  int plane, idx, nplane, j;

  if (pp->n_planes == 0)
    return NULL;  /* no known structure */

  plan_find_satellite(pp, ps->id, &plane, &idx);

  /* satellite not in plan */
  if (plane < 0)
    return NULL;

  nplane = plane + index;
  while (nplane<0)
    nplane += pp->n_planes;
  nplane = nplane % pp->n_planes;

  /* check that there are satellites in this plane */
  if (pp->sats_in_plane[nplane] == 0)
    return NULL;

  /* find closest member of this plane */
  closest = constellation_search(pc, pp->sat_ids[nplane][0]);
  range = distance(&(closest->x_C), &(ps->x_C));
  
  for (j=1; j<pp->sats_in_plane[nplane]; j++){
    current = constellation_search(pc, pp->sat_ids[nplane][j]);
    if (distance(&(current->x_C), &(ps->x_C)) < range){
      closest = current;
      range = distance(&(current->x_C), &(ps->x_C));
    }
  }
  return closest;
}

static void
plan_find_satellite(Plan *pp, int id, int *plane, int *index)
{
  int i, j;

  for (i=0; i<pp->n_planes; i++) {
    for (j=0; j<pp->sats_in_plane[i]; j++)
      if (pp->sat_ids[i][j] == id) {
	*plane = i;
	*index = j;
	return;
      }
  }
  
  /* not found! */
  *plane = -1;
  *index = -1;
}

