/*
 *****************************************************
 *
 *  SaVi by Robert Thurman (thurman@geom.umn.edu) and
 *          Patrick Worfolk (worfolk@alum.mit.edu).
 *
 *  Copyright (c) 1997 by The Geometry Center.
 *  This file is part of SaVi.  SaVi is free software;
 *  you can redistribute it and/or modify it only under
 *  the terms given in the file COPYRIGHT which you should
 *  have received along with this file.  SaVi may be
 *  obtained by emailing savi@geom.umn.edu.
 *
 * Modified:
 *  10/9/97  Patrick Worfolk  Extended for data output
 *
 *****************************************************
 *  
 * satellites.c
 *
 * routines for satellites in Geomview
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "globals.h"
#include "constants.h"
#include "gv_utils.h"
#include "sats.h"
#include "orbit_utils.h"
#include "savi.h"

static int satellites_on_flag = FALSE;
static int satellites_geom_exists = FALSE;
static int satellites_geom_needs_updating = FALSE;
static int use_fancy_satellite = FALSE;

/*
 * routine to display satellites
 */
char *
satellites_on_cmd(int argc, char *argv[])
{
  if (TRUE == satellites_on_flag) return EMPTY_str;
  satellites_on_flag = TRUE;
  satellites_geom_needs_updating = TRUE;

  if (TRUE == use_fancy_satellite) {
    transforms_needed |= (1 << SATELLITES);
    /* and send them too */
    gv_trans_list_create(get_constellation()->satellites);
  }
    

  gv_begin();
  write_satellites_geom(get_constellation());
  gv_end();

  return EMPTY_str;
}

/*
 * satellites_off_cmd()
 *
 * Routine to turn off display of satellites
 */
char *
satellites_off_cmd(int argc, char *argv[])
{
  if (FALSE == satellites_on_flag) return EMPTY_str;

  satellites_gv_delete();
  satellites_geom_needs_updating = FALSE;
  satellites_on_flag = FALSE;

  return EMPTY_str;
}


/*
 * routine to display fancy satellites
 */
char *
fancy_sat_on_cmd(int argc, char *argv[])
{
  if (TRUE == use_fancy_satellite) return EMPTY_str;

  use_fancy_satellite = TRUE;

  if (TRUE == satellites_on_flag) {
    transforms_needed |= (1 << SATELLITES);
    satellites_geom_needs_updating = TRUE;
    gv_begin();
    gv_trans_list_create(get_constellation()->satellites);
    write_satellites_geom(get_constellation());
    gv_end();
  }

  return EMPTY_str;
}

/*
 * satellites_off_cmd()
 *
 * Routine to turn off display of fancy satellites
 */
char *
fancy_sat_off_cmd(int argc, char *argv[])
{
  if (FALSE == use_fancy_satellite) return EMPTY_str;

  use_fancy_satellite = FALSE;

  if (TRUE == satellites_on_flag) {
    transforms_needed &= ~(1 << SATELLITES);
    satellites_geom_needs_updating = TRUE;
    gv_begin();
    write_satellites_geom(get_constellation());
    gv_end();
  }

  return EMPTY_str;
}


char *
satellites_set_cmd(int argc, char *argv[])
{
  Constellation *pconstellation = get_constellation();
  Satellite s;

  int n = atoi(argv[2]);
  unsigned int flag = atoi(argv[3]);

  s = constellation_search(pconstellation, n);
  if (s) {
    if (flag) s->can_display_satellite = 1;
    else s->can_display_satellite = 0;
    satellites_geom_needs_updating = TRUE;
  }

  return EMPTY_str;
}


/*
 * satellites_gv_delete
 *
 * routine to delete geomview satellites object
 */
void
satellites_gv_delete()
{
  if (TRUE == satellites_geom_exists) {
    gv_begin();
    gv_delete_geom("Satellites");
    gv_end();
    satellites_geom_exists = FALSE;
  }
}

/*
 * satellite_display
 *
 * Takes a satellite and displays it
 */
void
satellite_display(Satellite s, const Constellation *pconstellation)
{
  satellites_geom_needs_updating = TRUE;
}

/*
 * satellites_relocate
 *
 * Moves a satellite into its proper position
 */
void
satellites_relocate(const Constellation *pconstellation)
{
  /* fancy satellites automatically updated from transformation matrices */
  if (TRUE == use_fancy_satellite)
    return;

  satellites_geom_needs_updating = TRUE;
  write_satellites_geom(pconstellation);
}

/*
 * satellite_delete
 *
 * Takes a satellite and deletes it on the display
 */
void
satellite_delete(Satellite s)
{
  satellites_geom_needs_updating = TRUE;
}

/* #undef USE_GV_BINARY_FORMATS */
void
write_satellites_geom(const Constellation *pconstellation)
{
  Satellite s;
  Satellite_list sl;
  char buf[256];
  int i;
  static char hname[] = "satellite_h";
  static char special_hname[] = "special_satellite_h";

  if ((TRUE == satellites_on_flag) && 
      (TRUE == satellites_geom_needs_updating)) {
    
    satellites_geom_needs_updating = FALSE;

    /* if no sats then delete the Satellites gv object */
    if ( (pconstellation->n_view_sats == 0) &&
	(pconstellation->n_tag_sats == 0)) {
      satellites_gv_delete();
      return;
    }
	
    if (TRUE == use_fancy_satellite) {

      /* write out new list of all satellites */
      sl = pconstellation->satellites;
      gv_send("(geometry Satellites {LIST\n");
      while (NULL != sl) {
	s = sl->s;
	if (sl->s->can_display_satellite) {
	  sprintf(buf, "{INST transform:trans_%d geom:%s}\n",
		  s->id, s->tag ? special_hname : hname);
	  gv_send(buf);
	}
	sl = sl->next;
      }
      gv_send("})\n");
      
    } else {

      /* write out as a LIST of VECT */
      gv_send("(geometry Satellites {LIST\n");

      /* write out all viewable satellites as a single VECT */
      if (pconstellation->n_view_sats > 0) {
#ifdef USE_GV_BINARY_FORMATS
	unsigned int ibuf[3];
	float fbuf[3];
	float rgba[] = {0, 1, 0, 1}; /* green */
	unsigned short zero_one_buf[] = {0, 1};
	unsigned short *zero = zero_one_buf;
        unsigned short *one = zero_one_buf+1;

	/* convert stuff to big endian format */
	htons_buffer(zero_one_buf, 2);

	/* send VECT header */
	gv_send("{appearance {linewidth 7} VECT BINARY\n");
	/* send: n_view_sats n_view_sats 1 */
	ibuf[0] = ibuf[1] = pconstellation->n_view_sats;
	ibuf[2] = 1;
	htonl_buffer(ibuf,3);
	gv_send_binary_ints(ibuf, 3);
	/* send n_view_sats number of short 1s */
	for (i=0; i<pconstellation->n_view_sats; i++)
	  gv_send_binary_shorts(one,1);
	/* send short 1 followed by n_view_sats-1 number of short 0s */
	gv_send_binary_shorts(one,1);
	for (i=1; i<pconstellation->n_view_sats; i++)
	  gv_send_binary_shorts(zero,1);
	/* for each satellite to view, send 3 floats describing position */
	for (sl = pconstellation->satellites; NULL != sl; sl = sl->next) {
	  s = sl->s;
	  if (s->can_display_satellite) {
	    fbuf[0] = s->transform[3][0];
	    fbuf[1] = s->transform[3][1];
	    fbuf[2] = s->transform[3][2];
	    htonl_buffer((unsigned int *) fbuf,3);
	    gv_send_binary_floats(fbuf, 3);
	  }
	}
	/* send color */
	htonl_buffer((unsigned int *) rgba, 4);
	gv_send_binary_floats(rgba, 4);
	gv_send("}\n");
#else
	gv_send("{appearance {linewidth 7} VECT\n");
	sprintf(buf, "%d %d 1\n", pconstellation->n_view_sats,
		pconstellation->n_view_sats);
	gv_send(buf);
	for (i=0; i<pconstellation->n_view_sats; i++)
	  gv_send("1 ");
	gv_send("\n1 ");
	for (i=1; i<pconstellation->n_view_sats; i++)
	  gv_send("0 ");
	gv_send("\n");
	for (sl = pconstellation->satellites; NULL != sl; sl = sl->next) {
	  s = sl->s;
	  if (s->can_display_satellite) {
	    sprintf(buf, "%f %f %f\n", s->transform[3][0],
		    s->transform[3][1],
		    s->transform[3][2]);
	    gv_send(buf);
	  }
	}
	gv_send("0 1 0 1}\n");
#endif
      }

      /* write out any tagged satellites */
      if (pconstellation->n_tag_sats > 0) {
	sprintf(buf, "{appearance {linewidth 15} VECT\n%d %d 1\n",
		pconstellation->n_tag_sats, pconstellation->n_tag_sats);
	gv_send(buf);
	for (i=0; i<pconstellation->n_tag_sats; i++)
	  gv_send("1 ");
	gv_send("\n1 ");
	for (i=1; i<pconstellation->n_tag_sats; i++)
	  gv_send("0 ");
	gv_send("\n");
	for (sl = pconstellation->satellites; NULL != sl; sl = sl->next) {
	  s = sl->s;
	  if (s->tag) {
	    sprintf(buf, "%f %f %f\n", s->transform[3][0],
		    s->transform[3][1],
		    s->transform[3][2]);
	    gv_send(buf);
	  }
	}
	gv_send("1 1 0 1}\n");
      }

      /* close LIST object */
      gv_send("})\n");
    }
    
    satellites_geom_exists = TRUE;
  }

}

