/*
 *****************************************************
 *
 *  SaVi by Robert Thurman (thurman@geom.umn.edu) and
 *          Patrick Worfolk (worfolk@alum.mit.edu).
 *
 *  Copyright (c) 1997 Patrick Worfolk
 *  All rights reserved.
 *
 *****************************************************
 *  
 * sun.c
 *
 * Routines for creating fancy sun model
 *
 */
#include <stdio.h>
#include <math.h>

#include "gv_utils.h"
#include "utils.h"
#include "constants.h"
#include "globals.h"
#include "savi.h"
#include "sats.h"

static int sun_on_flag = FALSE;

/*
 * sun_on_cmd
 *
 * Adjusts lighting for day/night shadows
 * and draws line from central body towards Sun.
 *
 */
char *
sun_on_cmd(int argc, char *argv[])
{
  char *status = EMPTY_str;

  if (TRUE == sun_on_flag) return EMPTY_str;
  sun_on_flag = TRUE;

  /* gv_sendfile(SUN_LIGHTING); */
  sun_place(current_time(), (get_constellation())->pcb);

  return status;
}

/*
 * sun_off_cmd
 *
 * Resets to uniform lighting.
 */
char *
sun_off_cmd(int argc, char *argv[])
{
  if (FALSE == sun_on_flag) return EMPTY_str;

  gv_sendfile(AMBIENT_LIGHTING);
  gv_delete_geom("Sun");
  sun_on_flag = FALSE;

  return EMPTY_str;
}

/*
 * sun_place
 *
 * rotates the sun to correct position for time t
 *
 * Time 0 is determined to be an equinox with the
 * 0 longitude turned away from the Sun
 *
 */
void
sun_place(double t, const CentralBody *pcb)
{
  char buf[256];
  double theta, phi, x, y, z;

  if (FALSE == sun_on_flag) return;

  /* theta is the earth in the sun's orbit */
  theta = t / pcb->orbital_period * TWOPI + PI;

  /* phi is the angle above the equatorial plane */
  phi = pcb->tilt * sin(theta) * DEG_TO_RAD;

  /* position the light source at x, y, z */
  x = cos(theta)*cos(phi);
  y = sin(theta)*cos(phi);
  z = sin(phi);

  gv_begin();
  gv_send("(merge-ap World appearance { lighting { ambient .7 .7 .7 replacelights light { ");
  gv_send("color 1 1 1 position ");
  sprintf(buf, "%f %f %f", x, y, z);
  gv_send(buf);
  gv_send(" 0 location local } } } )");

  gv_send("(geometry Sun {VECT 1 2 1 2 1 ");
  gv_send(buf);
  sprintf(buf, " %f %f %f", 2*x, 2*y, 2*z);
  gv_send(buf);
  gv_send(" 1 1 0 1 } )");
  gv_end();
}


/*
 * sun_angle
 *
 * returns angle (in degrees) between point on central body under
 * satellite and the line to the sun
 *
 */
double
sun_angle(Satellite s, const CentralBody *pcb)
{
  return 90.0;
}
