/*
 *****************************************************
 *
 *  SaVi by Robert Thurman (thurman@geom.umn.edu) and
 *          Patrick Worfolk (worfolk@alum.mit.edu).
 *
 *  Copyright (c) 1997 by The Geometry Center.
 *  This file is part of SaVi.  SaVi is free software;
 *  you can redistribute it and/or modify it only under
 *  the terms given in the file COPYRIGHT which you should
 *  have received along with this file.  SaVi may be
 *  obtained by emailing savi@geom.umn.edu.
 *
 *****************************************************
 *  
 * utils.c
 *
 */
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
/* for select */
#include <unistd.h>
#include <sys/time.h>
#include <pwd.h>

#include "constants.h"
#include "utils.h"

/*
 * copyfile(stdout, "data.txt")
 *
 * Will copy the file data.txt to stdout.
 * Returns OK if succesful, else NOT_OK.
 */
int
copyfile(FILE *out_fp, const char filename[])
{
  FILE *in_fp;
  char buf[256];

  if (out_fp == NULL) return NOT_OK;
  
  if (NULL == (in_fp = fopen(filename, "r"))) return NOT_OK;

  while ( NULL != fgets(buf, 256, in_fp) )
    {
      fprintf(out_fp, "%s", buf);
    }
  fflush(out_fp);
  fclose(in_fp);

  return OK;
}

/*
 * forward over comments and any newlines and white space
 *
 */
void
forward_over_comments(FILE *fp)
{
  int c = getc(fp);

  while ((c == '\n') || (c == ' ') || (c == '\t') || (c == '#'))
    {
      if ( c == '#' )
	{
	  /* skip to end of line */
	  do {
	    c = getc(fp);
	  } while ((c != '\n') && (c != EOF));
	}
      c = getc(fp);
    }
  ungetc(c,fp);
}

/*
 * identity
 *
 * Fills in a 4x4 matrix to be the identity matrix
 */
void
identity(double m[4][4])
{
  int i,j;

  for (i=0; i<4; i++)
    for(j=0; j<4; j++)
      m[i][j] = ( (i==j) ? 1.0 : 0.0 );
}


/*
 * Return the norm of a CartesianCoordinate vector
 */
double
norm( CartesianCoordinates *px )
{
  return ( sqrt(px->x*px->x + px->y*px->y + px->z*px->z) );
}

/*
 * Normalizes a CartesianCoordinate vector to unit length
 */
void
normalize( CartesianCoordinates *px )
{
  double l = norm(px);

  if (l == 0.0) return;
  px->x /= l;
  px->y /= l;
  px->z /= l;
}

/*
 * Return the dot-product of two CartesianCoordinate vectors.
 */
double
dot( CartesianCoordinates *px, CartesianCoordinates *py)
{
  return ( px->x*py->x + px->y*py->y + px->z*py->z );
}


/*
 * The vector cross-product of two CartesianCoordinate vectors
 */
void 
cross_product( CartesianCoordinates *pucv, CartesianCoordinates *pu,
	      CartesianCoordinates *pv )
{
  pucv->x = (pu->y*pv->z-pu->z*pv->y);
  pucv->y = (pu->z*pv->x-pu->x*pv->z);
  pucv->z = (pu->x*pv->y-pu->y*pv->x);
}

/*
 * Linear combination of two CartesianCoordinate vectors.
 * 
 *         w=au+bv
 */
void 
linear_com(double a, CartesianCoordinates *u,  
	   double b, CartesianCoordinates *v, CartesianCoordinates *w)
{
  w->x = (a*u->x+b*v->x);
  w->y = (a*u->y+b*v->y);
  w->z = (a*u->z+b*v->z);
}

/*
 * Distance between two points given by CartesianCoordinate vectors u and v. 
 * I.e., norm(u-v). 
 */
double
distance(CartesianCoordinates *u, CartesianCoordinates *v)
{
  CartesianCoordinates w;
  linear_com(1.0, u, -1.0, v, &w);
  return(norm(&w));
}


void
print_vec(const char msg[], CartesianCoordinates *pv)
{
  fprintf(stderr,"%s = (%f,%f,%f)\n", msg, pv->x, pv->y, pv->z);
}

/* 
 * Rotate a vector by angle theta about the z-axis.
 *
 * The rotation matrix is
 *   
 *   cos(theta)    -sin(theta)    0
 *   sin(theta)     cos(theta)    0
 *       0              0         1
 *
 */
void
rotate_z (CartesianCoordinates *v_old, const double theta,
	  CartesianCoordinates *v_new)
{
  v_new->x = cos(theta) * v_old->x - sin(theta) * v_old->y;
  v_new->y = sin(theta) * v_old->x + cos(theta) * v_old->y;
  v_new->z = v_old->z;
}



/*
 * error_format
 *
 * prints a string using formatting to stderr
 *   - a newline is automatically appended
 */
void
error_format(const char format[], const char msg[])
{
  fprintf(stderr, format, msg);
  fprintf(stderr, "\n");
}


/*
 * error_and_exit
 *
 * prints an error msg and exits program with value 1
 */
void
error_and_exit(const char msg[])
{
  error_format("%s.  Exiting.", msg);
  exit(1);
}

/*
 * error
 *
 * prints a string as an error message
 */
void
error(const char msg[])
{
  error_format("%s.", msg);
}


void
millisleep(int ms)
{

  struct timeval tv;

  tv.tv_usec = (ms%1000) * 1000;
  tv.tv_sec = ms/1000;
  select(0, NULL, NULL, NULL, &tv);
}

/*
 * Returns string something like: Tue Oct  7 22:07:38 1997
 */
char *
get_the_time()
{
  time_t	clock;
  char		*the_time;

  clock = time((time_t *) NULL); 
  the_time = ctime(&clock);
  if ( (strlen(the_time) > 0) && (the_time[strlen(the_time)-1] == '\n'))
    /* get rid of trailer \n */
    the_time[strlen(the_time)-1] = '\0';
  return(the_time);
}

/*
 * Returns string something like: Patrick A. Worfolk
 * (taken from passwd file)
 */
char *
get_user_info()
{
  char		*name;
  struct passwd 	*p;

  if( (name = getlogin()) )
    p = getpwnam(name);
  else 
    p = getpwuid(getuid());
  return(p->pw_gecos);
}






